"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BigQueryDriver = void 0;
/* eslint-disable no-underscore-dangle */
const ramda_1 = __importDefault(require("ramda"));
const bigquery_1 = require("@google-cloud/bigquery");
const storage_1 = require("@google-cloud/storage");
const base_driver_1 = require("@cubejs-backend/base-driver");
const shared_1 = require("@cubejs-backend/shared");
const HydrationStream_1 = require("./HydrationStream");
class BigQueryDriver extends base_driver_1.BaseDriver {
    constructor(config = {}) {
        super();
        this.storage = null;
        this.bucket = null;
        this.options = {
            scopes: ['https://www.googleapis.com/auth/bigquery', 'https://www.googleapis.com/auth/drive'],
            projectId: process.env.CUBEJS_DB_BQ_PROJECT_ID,
            keyFilename: process.env.CUBEJS_DB_BQ_KEY_FILE,
            credentials: process.env.CUBEJS_DB_BQ_CREDENTIALS ?
                JSON.parse(Buffer.from(process.env.CUBEJS_DB_BQ_CREDENTIALS, 'base64').toString('utf8')) :
                undefined,
            exportBucket: shared_1.getEnv('dbExportBucket') || process.env.CUBEJS_DB_BQ_EXPORT_BUCKET,
            location: shared_1.getEnv('bigQueryLocation'),
            ...config,
            pollTimeout: (config.pollTimeout || shared_1.getEnv('dbPollTimeout') || shared_1.getEnv('dbQueryTimeout')) * 1000,
            pollMaxInterval: (config.pollMaxInterval || shared_1.getEnv('dbPollMaxInterval')) * 1000,
        };
        shared_1.getEnv('dbExportBucketType', {
            supported: ['gcp'],
        });
        this.bigquery = new bigquery_1.BigQuery(this.options);
        if (this.options.exportBucket) {
            this.storage = new storage_1.Storage(this.options);
            this.bucket = this.storage.bucket(this.options.exportBucket);
        }
    }
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 10;
    }
    static driverEnvVariables() {
        return ['CUBEJS_DB_BQ_PROJECT_ID', 'CUBEJS_DB_BQ_KEY_FILE'];
    }
    async testConnection() {
        await this.bigquery.query({
            query: 'SELECT ? AS number', params: ['1']
        });
    }
    readOnly() {
        return !!this.options.readOnly;
    }
    async query(query, values, options) {
        const data = await this.runQueryJob({
            query,
            params: values,
            parameterMode: 'positional',
            useLegacySql: false
        }, options);
        return (data[0] && data[0].map(row => ramda_1.default.map(value => (value && value.value && typeof value.value === 'string' ? value.value : value), row)));
    }
    async loadTablesForDataset(dataset) {
        try {
            const result = await dataset.query({
                query: `
        SELECT
          columns.column_name as ${this.quoteIdentifier('column_name')},
          columns.table_name as ${this.quoteIdentifier('table_name')},
          columns.table_schema as ${this.quoteIdentifier('table_schema')},
          columns.data_type as ${this.quoteIdentifier('data_type')}
        FROM INFORMATION_SCHEMA.COLUMNS
      `
            });
            if (result.length) {
                return ramda_1.default.reduce(this.informationColumnsSchemaReducer, {}, result[0]);
            }
            return [];
        }
        catch (e) {
            if (e.message.includes('Permission bigquery.tables.get denied on table')) {
                return {};
            }
            throw e;
        }
    }
    async tablesSchema() {
        const dataSets = await this.bigquery.getDatasets();
        const dataSetsColumns = await Promise.all(dataSets[0].map((dataSet) => this.loadTablesForDataset(dataSet)));
        return dataSetsColumns.reduce((prev, current) => Object.assign(prev, current), {});
    }
    async getTablesQuery(schemaName) {
        try {
            const dataSet = await this.bigquery.dataset(schemaName);
            if (!dataSet) {
                return [];
            }
            const [tables] = await this.bigquery.dataset(schemaName).getTables();
            return tables.map(t => ({ table_name: t.id }));
        }
        catch (e) {
            if (e.toString().indexOf('Not found')) {
                return [];
            }
            throw e;
        }
    }
    async tableColumnTypes(table) {
        const [schema, name] = table.split('.');
        const [bigQueryTable] = await this.bigquery.dataset(schema).table(name).getMetadata();
        return bigQueryTable.schema.fields.map((c) => ({ name: c.name, type: this.toGenericType(c.type) }));
    }
    async createSchemaIfNotExists(schemaName) {
        return this.bigquery.dataset(schemaName).get({ autoCreate: true });
    }
    async isUnloadSupported() {
        return this.bucket !== null;
    }
    async stream(query, values) {
        const stream = await this.bigquery.createQueryStream({
            query,
            params: values,
            parameterMode: 'positional',
            useLegacySql: false
        });
        const rowStream = new HydrationStream_1.HydrationStream();
        stream.pipe(rowStream);
        return {
            rowStream,
        };
    }
    async unload(table) {
        if (!this.bucket) {
            throw new Error('Unload is not configured');
        }
        const destination = this.bucket.file(`${table}-*.csv.gz`);
        const [schema, tableName] = table.split('.');
        const bigQueryTable = this.bigquery.dataset(schema).table(tableName);
        const [job] = await bigQueryTable.createExtractJob(destination, { format: 'CSV', gzip: true });
        await this.waitForJobResult(job, { table }, false);
        const [files] = await this.bucket.getFiles({ prefix: `${table}-` });
        const urls = await Promise.all(files.map(async (file) => {
            const [url] = await file.getSignedUrl({
                action: 'read',
                expires: new Date(new Date().getTime() + 60 * 60 * 1000)
            });
            return url;
        }));
        return { csvFile: urls };
    }
    async loadPreAggregationIntoTable(preAggregationTableName, loadSql, params, options) {
        const [dataSet, tableName] = preAggregationTableName.split('.');
        const bigQueryQuery = {
            query: loadSql,
            params,
            parameterMode: 'positional',
            destination: this.bigquery.dataset(dataSet).table(tableName),
            createDisposition: 'CREATE_IF_NEEDED',
            useLegacySql: false
        };
        return this.runQueryJob(bigQueryQuery, options, false);
    }
    async awaitForJobStatus(job, options, withResults) {
        const [result] = await job.getMetadata();
        if (result.status && result.status.state === 'DONE') {
            if (result.status.errorResult) {
                throw new Error(result.status.errorResult.message ?
                    result.status.errorResult.message :
                    JSON.stringify(result.status.errorResult));
            }
            this.reportQueryUsage(result.statistics, options);
        }
        else {
            return null;
        }
        return withResults ? job.getQueryResults() : true;
    }
    async runQueryJob(bigQueryQuery, options, withResults = true) {
        const [job] = await this.bigquery.createQueryJob(bigQueryQuery);
        return this.waitForJobResult(job, options, withResults);
    }
    async waitForJobResult(job, options, withResults) {
        const startedTime = Date.now();
        for (let i = 0; Date.now() - startedTime <= this.options.pollTimeout; i++) {
            const result = await this.awaitForJobStatus(job, options, withResults);
            if (result) {
                return result;
            }
            await shared_1.pausePromise(Math.min(this.options.pollMaxInterval, 200 * i));
        }
        await job.cancel();
        throw new Error(`BigQuery job timeout reached ${this.options.pollTimeout}ms`);
    }
    quoteIdentifier(identifier) {
        const nestedFields = identifier.split('.');
        return nestedFields.map(name => {
            if (name.match(/^[a-z0-9_]+$/)) {
                return name;
            }
            return `\`${identifier}\``;
        }).join('.');
    }
}
exports.BigQueryDriver = BigQueryDriver;
//# sourceMappingURL=BigQueryDriver.js.map